"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TokenTransfer = void 0;
const sdk_base_1 = require("@wormhole-foundation/sdk-base");
const sdk_definitions_1 = require("@wormhole-foundation/sdk-definitions");
const common_1 = require("../common");
const config_1 = require("../config");
const wormhole_1 = require("../wormhole");
const wormholeTransfer_1 = require("../wormholeTransfer");
class TokenTransfer {
    wh;
    // state machine tracker
    _state;
    // transfer details
    transfer;
    // txids, populated once transactions are submitted
    txids = [];
    // The corresponding vaa representing the TokenTransfer
    // on the source chain (if its been completed and finalized)
    vaas;
    constructor(wh, transfer) {
        this._state = wormholeTransfer_1.TransferState.Created;
        this.wh = wh;
        this.transfer = transfer;
    }
    getTransferState() {
        return this._state;
    }
    static async from(wh, from, timeout = 6000) {
        if ((0, sdk_definitions_1.isTokenTransferDetails)(from)) {
            await TokenTransfer.validateTransferDetails(wh, from);
            const fromChain = wh.getChain(from.from.chain);
            const toChain = wh.getChain(from.to.chain);
            // Apply hackery
            from = {
                ...from,
                ...(await TokenTransfer.destinationOverrides(fromChain, toChain, from)),
            };
            return new TokenTransfer(wh, from);
        }
        let tt;
        if ((0, sdk_definitions_1.isWormholeMessageId)(from)) {
            tt = await TokenTransfer.fromIdentifier(wh, from, timeout);
        }
        else if ((0, sdk_definitions_1.isTransactionIdentifier)(from)) {
            tt = await TokenTransfer.fromTransaction(wh, from, timeout);
        }
        else {
            throw new Error("Invalid `from` parameter for TokenTransfer");
        }
        await tt.fetchAttestation(timeout);
        return tt;
    }
    // init from the seq id
    static async fromIdentifier(wh, id, timeout) {
        const vaa = await TokenTransfer.getTransferVaa(wh, id, timeout);
        // Check if its a payload 3 targeted at a relayer on the destination chain
        const automatic = TokenTransfer.isAutomatic(wh, vaa);
        // TODO: the `from.address` here is a lie, but we don't
        // immediately have enough info to get the _correct_ one
        // TODO: grab at least the init tx from the api
        const from = { chain: vaa.emitterChain, address: vaa.emitterAddress };
        let { token, to } = vaa.payload;
        const { address, chain } = token;
        const decimals = await wh.getDecimals(token.chain, token.address);
        const rescale = (amt, decimals) => amt * 10n ** (decimals - 8n);
        const _amount = rescale(token.amount, decimals);
        let nativeGasAmount = 0n;
        if (automatic) {
            const relayerPayload = vaa.payload.payload;
            // TODO: how do we not have a payload for this yet
            // const type = relayerPayload.slice(0, 1);
            // const fee = rescale(encoding.bignum.decode(relayerPayload.slice(1, 33)), decimals);
            // Scale gas to account for truncated VAA amount
            nativeGasAmount = rescale(sdk_base_1.encoding.bignum.decode(relayerPayload.slice(33, 65)), decimals);
            // actual receiver
            const receiver = new sdk_definitions_1.UniversalAddress(relayerPayload.slice(65, 97));
            // Overwrite the receiver
            to = { chain: vaa.payload.to.chain, address: receiver };
        }
        const details = {
            token: { address, chain },
            amount: _amount,
            from,
            to,
            automatic,
            nativeGas: nativeGasAmount,
        };
        const tt = new TokenTransfer(wh, details);
        tt.vaas = [{ id: id, vaa }];
        tt._state = wormholeTransfer_1.TransferState.Attested;
        return tt;
    }
    static async fromTransaction(wh, from, timeout) {
        const msg = await TokenTransfer.getTransferMessage(wh.getChain(from.chain), from.txid, timeout);
        const tt = await TokenTransfer.fromIdentifier(wh, msg, timeout);
        tt.txids = [from];
        return tt;
    }
    // start the WormholeTransfer by submitting transactions to the source chain
    // returns a transaction hash
    async initiateTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this (eg: state == Created)
            1) get a token transfer transaction for the token bridge given the context
            2) sign it given the signer
            3) submit it to chain
            4) return transaction id
        */
        if (this._state !== wormholeTransfer_1.TransferState.Created)
            throw new Error("Invalid state transition in `start`");
        const fromChain = this.wh.getChain(this.transfer.from.chain);
        this.txids = await TokenTransfer.transfer(fromChain, this.transfer, signer);
        this._state = wormholeTransfer_1.TransferState.SourceInitiated;
        return this.txids.map(({ txid }) => txid);
    }
    // wait for the VAA to be ready
    // returns the sequence number
    async fetchAttestation(timeout) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Started)
            1) poll the api on an interval to check if the VAA is available
            2) Once available, pull the VAA and parse it
            3) return seq
        */
        if (this._state < wormholeTransfer_1.TransferState.SourceInitiated || this._state > wormholeTransfer_1.TransferState.Attested)
            throw new Error("Invalid state transition in `ready`");
        if (!this.vaas || this.vaas.length === 0) {
            if (this.txids.length === 0)
                throw new Error("No VAAs set and txids available to look them up");
            // TODO: assuming the _last_ transaction in the list will contain the msg id
            const txid = this.txids[this.txids.length - 1];
            const msgId = await TokenTransfer.getTransferMessage(this.wh.getChain(txid.chain), txid.txid, timeout);
            this.vaas = [{ id: msgId }];
        }
        for (const idx in this.vaas) {
            // Check if we already have the VAA
            if (this.vaas[idx].vaa)
                continue;
            this.vaas[idx].vaa = await TokenTransfer.getTransferVaa(this.wh, this.vaas[idx].id, timeout);
        }
        this._state = wormholeTransfer_1.TransferState.Attested;
        return this.vaas.map((vaa) => vaa.id);
    }
    // finish the WormholeTransfer by submitting transactions to the destination chain
    // returns a transaction hash
    async completeTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Ready)
            1) prepare the transactions and sign them given the signer
            2) submit the VAA and transactions on chain
            3) return txid of submission
        */
        if (this._state < wormholeTransfer_1.TransferState.Attested)
            throw new Error("Invalid state transition in `finish`. Be sure to call `fetchAttestation`.");
        if (!this.vaas)
            throw new Error("No VAA details available");
        // TODO: when do we get >1?
        const { vaa } = this.vaas[0];
        if (!vaa)
            throw new Error(`No VAA found for ${this.vaas[0].id.sequence}`);
        const toChain = this.wh.getChain(this.transfer.to.chain);
        const redeemTxids = await TokenTransfer.redeem(toChain, vaa, signer, this.transfer.automatic);
        this.txids.push(...redeemTxids);
        return redeemTxids.map(({ txid }) => txid);
    }
    // Async fn that produces status updates through an async generator
    // eventually producing a receipt
    // can be called repeatedly so the receipt is updated as it moves through the
    // steps of the transfer
    // Note: it should be possible to call this fn as many times
    // as it takes until the destination is finalized
    static async *track(wh, xfer, timeout = config_1.DEFAULT_TASK_TIMEOUT) {
        const start = Date.now();
        const leftover = (start, max) => Math.max(max - (Date.now() - start), 0);
        const _fromChain = wh.getChain(xfer.transfer.from.chain);
        const _toChain = wh.getChain(xfer.transfer.to.chain);
        const receipt = TokenTransfer.getReceipt(xfer);
        if (receipt.state === wormholeTransfer_1.TransferState.SourceInitiated) {
            if (receipt.originTxs.length === 0)
                throw "Invalid state transition";
            if (!receipt.attestation || !receipt.attestation?.emitter) {
                const initTx = receipt.originTxs[receipt.originTxs.length - 1];
                const xfermsg = await TokenTransfer.getTransferMessage(_fromChain, initTx.txid, leftover(start, timeout));
                receipt.attestation = { ...xfermsg };
                receipt.state = wormholeTransfer_1.TransferState.SourceFinalized;
                yield receipt.state;
            }
        }
        let vaa = undefined;
        if (receipt.state == wormholeTransfer_1.TransferState.SourceFinalized) {
            if (!receipt.attestation || !receipt.attestation.emitter)
                throw "Invalid state transition";
            vaa = xfer.vaas && xfer.vaas.length > 0 && xfer.vaas[0].vaa ? xfer.vaas[0].vaa : undefined;
            // we need to get the attestation so we can deliver it
            // we can use the message id we parsed out of the logs, if we have them
            // or try to fetch it from the last origin transaction
            if (!vaa) {
                vaa = await TokenTransfer.getTransferVaa(wh, { ...receipt.attestation }, leftover(start, timeout));
                receipt.state = wormholeTransfer_1.TransferState.Attested;
                yield receipt.state;
            }
        }
        if (receipt.state == wormholeTransfer_1.TransferState.Attested) {
            if (!receipt.attestation || !vaa)
                throw "Invalid state transition";
            // First try to grab the tx status from the API
            // Note: this requires a subsequent async step on the backend
            // to have the dest txid populated, so it may be delayed by some time
            const txStatus = await wh.getTransactionStatus(receipt.attestation, leftover(start, timeout));
            if (!txStatus)
                return receipt;
            if (txStatus.globalTx?.destinationTx?.txHash) {
                const { chainId, txHash } = txStatus.globalTx.destinationTx;
                receipt.destinationTxs = [
                    {
                        chain: (0, sdk_base_1.toChain)(chainId),
                        txid: txHash,
                    },
                ];
                receipt.state = wormholeTransfer_1.TransferState.DestinationFinalized;
                yield receipt.state;
            }
            // Fall back to asking the destination chain if this VAA has been redeemed
            if ((await TokenTransfer.isTransferComplete(_toChain, vaa), leftover(start, timeout))) {
                receipt.state = wormholeTransfer_1.TransferState.DestinationFinalized;
                yield receipt.state;
            }
        }
        return receipt;
    }
    // Static method to perform the transfer so a custom RPC may be used
    // Note: this assumes the transfer has already been validated with `validateTransfer`
    static async transfer(fromChain, transfer, signer) {
        const senderAddress = (0, sdk_definitions_1.toNative)(signer.chain(), signer.address());
        const token = transfer.token === "native"
            ? "native"
            : transfer.token.address;
        let xfer;
        if (transfer.automatic) {
            const tb = await fromChain.getAutomaticTokenBridge();
            xfer = tb.transfer(senderAddress, transfer.to, token, transfer.amount, transfer.nativeGas);
        }
        else {
            const tb = await fromChain.getTokenBridge();
            xfer = tb.transfer(senderAddress, transfer.to, token, transfer.amount, transfer.payload);
        }
        return (0, common_1.signSendWait)(fromChain, xfer, signer);
    }
    // Static method to allow passing a custom RPC
    static async redeem(toChain, vaa, signer, automatic) {
        const signerAddress = (0, sdk_definitions_1.toNative)(signer.chain(), signer.address());
        let xfer;
        if (automatic) {
            if (vaa.payloadName === "Transfer")
                throw new Error("VAA is a simple transfer but expected Payload for automatic delivery");
            const tb = await toChain.getAutomaticTokenBridge();
            xfer = tb.redeem(signerAddress, vaa);
        }
        else {
            const tb = await toChain.getTokenBridge();
            xfer = tb.redeem(signerAddress, vaa);
        }
        return (0, common_1.signSendWait)(toChain, xfer, signer);
    }
    static async isTransferComplete(toChain, vaa) {
        const tb = await toChain.getTokenBridge();
        return tb.isTransferCompleted(vaa);
    }
    static async getTransferMessage(chain, txid, timeout) {
        // A Single wormhole message will be returned for a standard token transfer
        const whm = await wormhole_1.Wormhole.parseMessageFromTx(chain, txid, timeout);
        if (whm.length !== 1)
            throw new Error("Expected a single Wormhole Message, got: " + whm.length);
        return whm[0];
    }
    static async getTransferVaa(wh, key, timeout) {
        const vaa = typeof key === "string"
            ? await wh.getVaaByTxHash(key, sdk_definitions_1.TokenBridge.getTransferDiscriminator(), timeout)
            : await wh.getVaa(key, sdk_definitions_1.TokenBridge.getTransferDiscriminator(), timeout);
        if (!vaa)
            throw new Error(`No VAA available after retries exhausted`);
        return vaa;
    }
    // Lookup the token id for the destination chain given the source chain
    // and token id
    static async lookupDestinationToken(srcChain, dstChain, token) {
        // that will be minted when the transfer is redeemed
        let lookup;
        if (token === "native") {
            // if native, get the wrapped asset id
            lookup = await srcChain.getNativeWrappedTokenId();
        }
        else {
            try {
                const tb = await srcChain.getTokenBridge();
                // otherwise, check to see if it is a wrapped token locally
                lookup = await tb.getOriginalAsset(token.address);
            }
            catch {
                // not a from-chain native wormhole-wrapped one
                lookup = token;
            }
        }
        // if the token id is actually native to the destination, return it
        if (lookup.chain === dstChain.chain) {
            const { chain, address } = lookup;
            return { chain, address: address.toNative(chain) };
        }
        // otherwise, figure out what the token address representing the wormhole-wrapped token we're transferring
        const dstTb = await dstChain.getTokenBridge();
        const dstAddress = await dstTb.getWrappedAsset(lookup);
        return { chain: dstChain.chain, address: dstAddress };
    }
    static isAutomatic(wh, vaa) {
        // Check if its a payload 3 targeted at a relayer on the destination chain
        const { chain, address } = vaa.payload.to;
        const { tokenBridgeRelayer } = wh.config.chains[chain].contracts;
        const relayerAddress = tokenBridgeRelayer
            ? (0, sdk_definitions_1.nativeChainAddress)(chain, tokenBridgeRelayer).address.toUniversalAddress()
            : null;
        return (vaa.payloadName === "TransferWithPayload" &&
            !!relayerAddress &&
            address.equals(relayerAddress));
    }
    static async validateTransferDetails(wh, transfer) {
        if (transfer.payload && transfer.automatic)
            throw new Error("Payload with automatic delivery is not supported");
        if (transfer.nativeGas && !transfer.automatic)
            throw new Error("Gas Dropoff is only supported for automatic transfers");
        const fromChain = wh.getChain(transfer.from.chain);
        const toChain = wh.getChain(transfer.to.chain);
        if (!fromChain.supportsTokenBridge())
            throw new Error(`Token Bridge not supported on ${transfer.from.chain}`);
        if (!toChain.supportsTokenBridge())
            throw new Error(`Token Bridge not supported on ${transfer.to.chain}`);
        if (transfer.automatic && !fromChain.supportsAutomaticTokenBridge())
            throw new Error(`Automatic Token Bridge not supported on ${transfer.from.chain}`);
        if (transfer.amount === 0n)
            throw new Error("Amount cannot be 0");
    }
    static async quoteTransfer(srcChain, dstChain, transfer) {
        const dstToken = await this.lookupDestinationToken(srcChain, dstChain, transfer.token);
        const srcToken = transfer.token === "native" ? await srcChain.getNativeWrappedTokenId() : transfer.token;
        const dstDecimals = await dstChain.getDecimals(dstToken.address);
        const srcDecimals = await srcChain.getDecimals(srcToken.address);
        // truncate the amount for over-the-wire max representation
        const truncate = (amt, decimals) => amt / 10n ** (decimals - 8n);
        // scale by number of decimals, pair with truncate
        const scale = (amt, decimals) => amt * 10n ** (decimals - 8n);
        // dedust for source using the number of decimals on the source chain
        // but scale back to original decimals
        const dedust = (amt) => scale(truncate(amt, srcDecimals), srcDecimals);
        // truncate and rescale to fit destination token
        const mintable = (amt) => scale(truncate(amt, srcDecimals), dstDecimals);
        // dedust for transfer over the bridge which truncates to 8 decimals
        let srcAmount = dedust(transfer.amount);
        let dstAmount = mintable(transfer.amount);
        if (!transfer.automatic) {
            return {
                sourceToken: { token: srcToken, amount: srcAmount },
                destinationToken: { token: dstToken, amount: dstAmount },
            };
        }
        // Otherwise automatic
        // If a native gas dropoff is requested, remove that from the amount they'll get
        const _nativeGas = transfer.nativeGas ? mintable(transfer.nativeGas) : 0n;
        dstAmount -= _nativeGas;
        // The fee is also removed from the amount transferred
        // quoted on the source chain
        const stb = await srcChain.getAutomaticTokenBridge();
        const fee = await stb.getRelayerFee(transfer.from.address, transfer.to, srcToken.address);
        dstAmount -= mintable(fee);
        // The expected destination gas can be pulled from the destination token bridge
        let destinationNativeGas = 0n;
        if (transfer.nativeGas) {
            const dtb = await dstChain.getAutomaticTokenBridge();
            destinationNativeGas = await dtb.nativeTokenAmount(dstToken.address, _nativeGas);
        }
        return {
            sourceToken: {
                token: srcToken,
                amount: srcAmount,
            },
            destinationToken: { token: dstToken, amount: dstAmount },
            relayFee: { token: srcToken, amount: fee },
            destinationNativeGas,
        };
    }
    static async destinationOverrides(srcChain, dstChain, transfer) {
        const _transfer = { ...transfer };
        // Bit of (temporary) hackery until solana contracts support being
        // sent a VAA with the primary address
        // Note: Do _not_ override if automatic or if the destination token is native
        // gas token
        if (transfer.to.chain === "Solana" && !_transfer.automatic) {
            const destinationToken = await TokenTransfer.lookupDestinationToken(srcChain, dstChain, _transfer.token);
            // TODO: If the token is native, no need to overwrite the destination address check for native
            //if (!destinationToken.address.equals((await dstChain.getNativeWrappedTokenId()).address))
            _transfer.to = await dstChain.getTokenAccount(_transfer.to.address, destinationToken.address);
        }
        if (_transfer.to.chain === "Sei") {
            if (_transfer.to.chain === "Sei" && _transfer.payload)
                throw new Error("Arbitrary payloads unsupported for Sei");
            // For sei, we reserve the payload for a token transfer through the sei bridge.
            _transfer.payload = sdk_base_1.encoding.bytes.encode(JSON.stringify({
                basic_recipient: {
                    recipient: sdk_base_1.encoding.b64.encode(_transfer.to.address.toString()),
                },
            }));
            _transfer.to = (0, sdk_definitions_1.nativeChainAddress)(_transfer.to.chain, dstChain.config.contracts.translator);
        }
        return _transfer;
    }
    static getReceipt(xfer) {
        const att = xfer.vaas && xfer.vaas.length > 0 ? xfer.vaas[0] : undefined;
        const receipt = {
            from: xfer.transfer.from.chain,
            to: xfer.transfer.to.chain,
            state: wormholeTransfer_1.TransferState.SourceInitiated,
            originTxs: xfer.txids.filter((txid) => txid.chain === xfer.transfer.from.chain),
            destinationTxs: xfer.txids.filter((txid) => txid.chain === xfer.transfer.to.chain),
            attestation: att && att.id.emitter ? { ...att.id } : undefined,
        };
        if (receipt.originTxs.length > 0)
            receipt.state = wormholeTransfer_1.TransferState.SourceInitiated;
        if (receipt.attestation && receipt.attestation.emitter)
            receipt.state = wormholeTransfer_1.TransferState.Attested;
        if (receipt.destinationTxs.length > 0)
            receipt.state = wormholeTransfer_1.TransferState.DestinationInitiated;
        return receipt;
    }
}
exports.TokenTransfer = TokenTransfer;
//# sourceMappingURL=tokenTransfer.js.map