import { circle, encoding } from "@wormhole-foundation/sdk-base";
import { deserializeCircleMessage, isCircleMessageId, isCircleTransferDetails, isTransactionIdentifier, isWormholeMessageId, nativeChainAddress, } from "@wormhole-foundation/sdk-definitions";
import { signSendWait } from "../common";
import { DEFAULT_TASK_TIMEOUT } from "../config";
import { Wormhole } from "../wormhole";
import { TransferState } from "../wormholeTransfer";
export class CircleTransfer {
    wh;
    // state machine tracker
    _state;
    // transfer details
    transfer;
    // Populated after Initialized
    txids = [];
    // Populated if !automatic and after initialized
    circleAttestations;
    // Populated if automatic and after initialized
    vaas;
    constructor(wh, transfer) {
        this._state = TransferState.Created;
        this.wh = wh;
        this.transfer = transfer;
    }
    getTransferState() {
        return this._state;
    }
    static async from(wh, from, timeout = DEFAULT_TASK_TIMEOUT) {
        // This is a new transfer, just return the object
        if (isCircleTransferDetails(from)) {
            return new CircleTransfer(wh, from);
        }
        // This is an existing transfer, fetch the details
        let tt;
        if (isWormholeMessageId(from)) {
            tt = await CircleTransfer.fromWormholeMessageId(wh, from, timeout);
        }
        else if (isTransactionIdentifier(from)) {
            tt = await CircleTransfer.fromTransaction(wh, from, timeout);
        }
        else if (isCircleMessageId(from)) {
            tt = await CircleTransfer.fromCircleMessage(wh, from, timeout);
        }
        else {
            throw new Error("Invalid `from` parameter for CircleTransfer");
        }
        await tt.fetchAttestation(timeout);
        return tt;
    }
    // init from the seq id
    static async fromWormholeMessageId(wh, from, timeout) {
        const { chain, emitter } = from;
        const vaa = await CircleTransfer.getTransferVaa(wh, from);
        const rcvAddress = vaa.payload.mintRecipient;
        const rcvChain = circle.toCircleChain(vaa.payload.targetDomain);
        // Check if its a payload 3 targeted at a relayer on the destination chain
        const { wormholeRelayer } = wh.config.chains[rcvChain].contracts.cctp;
        let automatic = false;
        if (wormholeRelayer) {
            const relayerAddress = nativeChainAddress(chain, wormholeRelayer).address.toUniversalAddress();
            automatic = vaa.payloadName === "TransferRelay" && rcvAddress.equals(relayerAddress);
        }
        const details = {
            from: nativeChainAddress(from.chain, vaa.payload.caller),
            to: nativeChainAddress(rcvChain, rcvAddress),
            amount: vaa.payload.token.amount,
            automatic,
        };
        const tt = new CircleTransfer(wh, details);
        tt.vaas = [{ id: { emitter, sequence: vaa.sequence, chain: chain }, vaa }];
        tt._state = TransferState.Attested;
        return tt;
    }
    static async fromCircleMessage(wh, message, timeout) {
        const [msg, hash] = deserializeCircleMessage(encoding.hex.decode(message));
        const { payload: burnMessage } = msg;
        const xferSender = burnMessage.messageSender;
        const xferReceiver = burnMessage.mintRecipient;
        const sendChain = circle.toCircleChain(msg.sourceDomain);
        const rcvChain = circle.toCircleChain(msg.destinationDomain);
        const details = {
            from: nativeChainAddress(sendChain, xferSender),
            to: nativeChainAddress(rcvChain, xferReceiver),
            amount: burnMessage.amount,
            automatic: false,
        };
        const xfer = new CircleTransfer(wh, details);
        xfer.circleAttestations = [{ id: { hash }, message: msg }];
        xfer._state = TransferState.SourceInitiated;
        return xfer;
    }
    // init from source tx hash
    static async fromTransaction(wh, from, timeout) {
        const { chain, txid } = from;
        const originChain = wh.getChain(chain);
        // First try to parse out a WormholeMessage
        // If we get one or more, we assume its a Wormhole attested
        // transfer
        const msgIds = await originChain.parseTransaction(txid);
        // If we found a VAA message, use it
        let ct;
        if (msgIds.length > 0) {
            ct = await CircleTransfer.fromWormholeMessageId(wh, msgIds[0], timeout);
        }
        else {
            // Otherwise try to parse out a circle message
            const cb = await originChain.getCircleBridge();
            const circleMessage = await cb.parseTransactionDetails(txid);
            const details = {
                ...circleMessage,
                // Note: assuming automatic is false since we didn't find a VAA
                automatic: false,
            };
            ct = new CircleTransfer(wh, details);
            ct.circleAttestations = [{ id: circleMessage.id, message: circleMessage.message }];
        }
        ct._state = TransferState.SourceInitiated;
        ct.txids = [from];
        return ct;
    }
    // start the WormholeTransfer by submitting transactions to the source chain
    // returns a transaction hash
    async initiateTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this (eg: state == Created)
            1) get a token transfer transaction for the token bridge given the context
            2) sign it given the signer
            3) submit it to chain
            4) return transaction id
        */
        if (this._state !== TransferState.Created)
            throw new Error("Invalid state transition in `start`");
        const fromChain = this.wh.getChain(this.transfer.from.chain);
        let xfer;
        if (this.transfer.automatic) {
            const cr = await fromChain.getAutomaticCircleBridge();
            xfer = cr.transfer(this.transfer.from.address, { chain: this.transfer.to.chain, address: this.transfer.to.address }, this.transfer.amount, this.transfer.nativeGas);
        }
        else {
            const cb = await fromChain.getCircleBridge();
            xfer = cb.transfer(this.transfer.from.address, { chain: this.transfer.to.chain, address: this.transfer.to.address }, this.transfer.amount);
        }
        this.txids = await signSendWait(fromChain, xfer, signer);
        this._state = TransferState.SourceInitiated;
        return this.txids.map(({ txid }) => txid);
    }
    async _fetchWormholeAttestation(timeout) {
        if (!this.vaas || this.vaas.length == 0)
            throw new Error("No VAA details available");
        // Check if we already have the VAA
        for (const idx in this.vaas) {
            // already got it
            if (this.vaas[idx].vaa)
                continue;
            this.vaas[idx].vaa = await CircleTransfer.getTransferVaa(this.wh, this.vaas[idx].id);
        }
        return this.vaas.map((v) => v.id);
    }
    async _fetchCircleAttestation(timeout) {
        if (!this.circleAttestations || this.circleAttestations.length == 0) {
            // If we dont have any circle attestations yet, we need to start by
            // fetching the transaction details from the source chain
            if (this.txids.length === 0)
                throw new Error("No circle attestations or transactions to fetch");
            // The last tx should be the circle transfer, its possible there was
            // a contract spend approval transaction
            const txid = this.txids[this.txids?.length - 1];
            const fromChain = this.wh.getChain(this.transfer.from.chain);
            const cb = await fromChain.getCircleBridge();
            const circleMessage = await cb.parseTransactionDetails(txid.txid);
            this.circleAttestations = [{ id: circleMessage.id, message: circleMessage.message }];
        }
        for (const idx in this.circleAttestations) {
            const ca = this.circleAttestations[idx];
            if (ca.attestation)
                continue; // already got it
            const attestation = await this.wh.getCircleAttestation(ca.id.hash, timeout);
            if (attestation === null)
                throw new Error("No attestation available after timeout exhausted");
            this.circleAttestations[idx].attestation = attestation;
        }
        return this.circleAttestations.map((v) => v.id);
    }
    // wait for the VAA to be ready
    // returns the sequence number
    async fetchAttestation(timeout) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Started)
            1) poll the api on an interval to check if the VAA is available
            2) Once available, pull the VAA and parse it
            3) return seq
        */
        if (this._state < TransferState.SourceInitiated || this._state > TransferState.Attested)
            throw new Error("Invalid state transition in `fetchAttestation`");
        const ids = this.transfer.automatic
            ? await this._fetchWormholeAttestation(timeout)
            : await this._fetchCircleAttestation(timeout);
        this._state = TransferState.Attested;
        return ids;
    }
    // finish the WormholeTransfer by submitting transactions to the destination chain
    // returns a transaction hash
    async completeTransfer(signer) {
        /*
            0) check that the current `state` is valid to call this  (eg: state == Ready)
            1) prepare the transactions and sign them given the signer
            2) submit the VAA and transactions on chain
            3) return txid of submission
        */
        if (this._state < TransferState.Attested)
            throw new Error("Invalid state transition in `finish`");
        // If its automatic, this does not need to be called
        if (this.transfer.automatic) {
            if (!this.vaas)
                throw new Error("No VAA details available");
            if (this.vaas.length > 1)
                throw new Error(`Expected a VAA, found ${this.vaas.length}`);
            const { vaa } = this.vaas[0];
            if (!vaa)
                throw new Error("No VAA found");
            //const tb = await toChain.getAutomaticCircleBridge();
            //const xfer = tb.redeem(vaa);
            //const txids = await signSendWait(toChain, xfer, signer);
            throw new Error("No method to redeem auto circle bridge tx (yet)");
        }
        if (!this.circleAttestations)
            throw new Error("No Circle Attestations found");
        if (this.circleAttestations.length > 1)
            throw new Error(`Expected a single circle attestation, found ${this.circleAttestations.length}`);
        const toChain = this.wh.getChain(this.transfer.to.chain);
        const { id, message, attestation } = this.circleAttestations[0];
        if (!attestation)
            throw new Error(`No Circle Attestation for ${id.hash}`);
        const tb = await toChain.getCircleBridge();
        const xfer = tb.redeem(this.transfer.to.address, message, attestation);
        const txids = await signSendWait(toChain, xfer, signer);
        this.txids?.push(...txids);
        return txids.map(({ txid }) => txid);
    }
    static async quoteTransfer(srcChain, dstChain, transfer) {
        const dstUsdcAddress = circle.usdcContract.get(dstChain.network, dstChain.chain);
        if (!dstUsdcAddress)
            throw "Invalid transfer, no USDC contract on destination";
        const srcUsdcAddress = circle.usdcContract.get(srcChain.network, srcChain.chain);
        if (!srcUsdcAddress)
            throw "Invalid transfer, no USDC contract on source";
        const dstToken = Wormhole.chainAddress(dstChain.chain, dstUsdcAddress);
        const srcToken = Wormhole.chainAddress(srcChain.chain, srcUsdcAddress);
        if (!transfer.automatic) {
            return {
                sourceToken: { token: srcToken, amount: transfer.amount },
                destinationToken: { token: dstToken, amount: transfer.amount },
            };
        }
        // Otherwise automatic
        let dstAmount = transfer.amount;
        // If a native gas dropoff is requested, remove that from the amount they'll get
        const _nativeGas = transfer.nativeGas ? transfer.nativeGas : 0n;
        dstAmount -= _nativeGas;
        // The fee is also removed from the amount transferred
        // quoted on the source chain
        const stb = await srcChain.getAutomaticCircleBridge();
        const fee = await stb.getRelayerFee(dstChain.chain);
        dstAmount -= fee;
        // The expected destination gas can be pulled from the destination token bridge
        let destinationNativeGas = 0n;
        if (transfer.nativeGas) {
            const dtb = await dstChain.getAutomaticTokenBridge();
            destinationNativeGas = await dtb.nativeTokenAmount(dstToken.address, _nativeGas);
        }
        return {
            sourceToken: {
                token: srcToken,
                amount: transfer.amount,
            },
            destinationToken: { token: dstToken, amount: dstAmount },
            relayFee: { token: srcToken, amount: fee },
            destinationNativeGas,
        };
    }
    static async getTransferVaa(wh, wormholeMessageId, timeout) {
        const vaa = await wh.getVaa(wormholeMessageId, "AutomaticCircleBridge:TransferRelay", timeout);
        if (!vaa)
            throw new Error(`No VAA available after timeout exhausted`);
        return vaa;
    }
}
//# sourceMappingURL=cctpTransfer.js.map